unit unitHighlighterImpl;
(* see readme.txt *)

interface

{$ifdef VER160}
procedure IDERegister;
{$else}
procedure Register;
{$endif}
procedure Cleanup;

implementation

uses  ToolsApi
    , SysUtils
    , Classes
    , Windows     { for OutputDebugString() }
    ;

type
  { a basic class implementing a highlighter and providing the basic IDE
    integration functions.
    }
  TCustomHighlighter = class( TObject, IInterface, IOTANotifier, IOTAHighlighter )
  protected
    FName, FIDString: String;
    FHighlighterIndex: Integer;
  protected // IInterface
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
  protected // IOTANotifier
    procedure AfterSave; virtual;
    procedure BeforeSave; virtual;
    procedure Destroyed; virtual;
    procedure Modified; virtual;
  protected // IOTAHighlighter
    function GetIDString: string; virtual;
    function GetName: string; virtual;
    procedure Tokenize(StartClass: TOTALineClass; LineBuf: POTAEdChar;
      LineBufLen: TOTALineSize; HighlightCodes: POTASyntaxCode); virtual;
    function TokenizeLineClass(StartClass: TOTALineClass;
      LineBuf: POTAEdChar; LineBufLen: TOTALineSize): TOTALineClass; virtual;
  protected // helpers
    procedure AddMessage( const sz: String ); virtual;
    function  GetLineText( p: PChar; len: Integer ): String; virtual;
    function  GetHighlighterList: String;
    function  FindHighlighter( const AID: String ): IOTAHighlighter;
    procedure RemoveHighlighter( const AID: String );
    procedure InstallHighlighter( AHighlighter: IOTAHighlighter );
  public
    constructor Create( const AName, AID: String );
    destructor  Destroy; override;
    end;

  { this class can be used to create a highlighter that is mean to replace
    one of the stock highlighters via delegation.
    NOTE: While the delegation works fine I have not been able to have
          it operate properly when the package is unloaded (when using
          a package) or have delphi shut down without issues (as a dll).
          So, good luck and let me know if you happen to know what the
          issue is.
    }
  TDelegateHighlighter = class( TCustomHighlighter )
  protected
    FDelegate: IOTAHighlighter;
  protected
    procedure Tokenize(StartClass: TOTALineClass; LineBuf: POTAEdChar;
      LineBufLen: TOTALineSize; HighlightCodes: POTASyntaxCode); override;
    function TokenizeLineClass(StartClass: TOTALineClass;
      LineBuf: POTAEdChar; LineBufLen: TOTALineSize): TOTALineClass; override;
  public
    constructor Create( const AName, AID: String );
    destructor  Destroy; override;
    end;

  { This is an alternate to the delegate. It aggregates and existing
    highlighter to create a new highlighter. It doesn't permit you to replace
    an existing highlighter (i.e. not as nice) but it allows you to leverage
    the existing highlighter to do most of the grunt work for you.
    }
  TAggregateHighlighter = class( TCustomHighlighter )
  protected
    FAggregate: IOTAHighlighter;
    FAggregateID: String;
  protected
    procedure Tokenize(StartClass: TOTALineClass; LineBuf: POTAEdChar;
      LineBufLen: TOTALineSize; HighlightCodes: POTASyntaxCode); override;
    function TokenizeLineClass(StartClass: TOTALineClass;
      LineBuf: POTAEdChar; LineBufLen: TOTALineSize): TOTALineClass; override;
  public
    constructor Create( const AName, AID, AAggregateID: String );
    destructor  Destroy; override;
    end;

type
  { a simple example that make it so that the following symbols are treated
    differently [, ], =
    }
//  TIniHighlighter = class( TCustomHighlighter )
//  TIniHighlighter = class( TDelegateHighlighter )
  TIniHighlighter = class( TAggregateHighlighter )
  public
    constructor Create;
  protected
    procedure Tokenize(StartClass: TOTALineClass; LineBuf: POTAEdChar;
      LineBufLen: TOTALineSize; HighlightCodes: POTASyntaxCode); override;
    end;


{ ------------------------ private globals --------------------------- }
var
  highlighters: array[0..0] of TCustomHighlighter = (nil);

{ ----------------- registration/cleanup/support --------------------- }

procedure DebugString( const sz: String );
begin
  Windows.OutputDebugString( pchar(sz) );
end;

{$ifdef VER160}
type
  { When trying to get the highlighter into D8.Net I struggled with
    a series of Interface not supported errors and for the life of
    me I was unable to figure out what was going on. It wasn't until
    I started looking at the docs for the C# tool api that I found
    out that all the services may not be available (which is what I
    had figured, just didn't know why) at the point where we are
    loaded.

    If you are making the expert in C#/D8.Net you can add a notification
    event to the 'BorlandIDE' object and get called when (and if) that
    service is loaded. There is nothing in the new ToolsAPI that would
    suggest that such a thing is possible for our situation (native
    wizard).
    }
  TProcedure = procedure;
  TWaitThread = class( TThread )
  private
    FProcedure: TProcedure;
    FWait: Integer;
    procedure DoProcedure;
  public
    constructor Create( AProcedure: TProcedure; AWait: Integer );
  protected
    procedure Execute; override;
    end;

{ TWaitThread }

constructor TWaitThread.Create( AProcedure: TProcedure; AWait: Integer );
begin
  inherited Create( true );
  FProcedure := AProcedure;
  FWait := AWait;
  FreeOnTerminate := true;
  Priority := tpLower;
  Resume;
end;

procedure TWaitThread.DoProcedure;
begin
  FProcedure;
end;

procedure TWaitThread.Execute;
begin
  Sleep( FWait * 1000 );
  Synchronize( DoProcedure );
end;
{$endif}

procedure RealRegister;
begin
  highlighters[0] := TIniHighlighter.Create;
end;

{$ifdef VER160}
procedure IDERegister;
begin
  TWaitThread.Create( RealRegister, 5 );
end;
{$else}
procedure Register;
begin
  RealRegister;
end;
{$endif}

procedure Cleanup;
var
  i: Integer;
begin
  for i := Low( highlighters ) to High( highlighters ) do
    try
      FreeAndNil( highlighters[i] );
    except
      OutputDebugString( pchar('Cleanup failed on highlighter: ' + IntToStr( i )) );
      end;
end;

function  GetIDEInterface( const iid: TGUID; out service ): Boolean;
begin
  assert( BorlandIDEServices <> nil );
{$ifdef VER160}
  {$message hint 'Using BorlandIDEServices.GetService()'}
  Result := BorlandIDEServices.GetService( iid, service );
{$else}
  Result := Supports( BorlandIDEServices, iid, service );
{$endif}
end;

function  GetIDEMessageService: IOTAMessageServices;
begin
  try
    if not GetIDEInterface( IOTAMessageServices, Result ) then
      raise Exception.Create( 'Interface not supported.' );
  except
    DebugString( 'Failed to get message services.'#13 +
                 Exception(ExceptObject).Message );
    raise;
    end;
end;

function  GetIDEHighlighterServices: IOTAHighlightServices;
begin
  try
    if not GetIDEInterface( IOTAHighlightServices, Result ) then
      raise Exception.Create( 'Interface not supported.' );
  except
    DebugString( 'Failed to get highlighter services.'#13 +
                 Exception(ExceptObject).Message );
    raise;
    end;
end;


{ ---------------------------- classes ------------------------------- }
{ TCustomHighlighter }

// stubs
procedure TCustomHighlighter.AfterSave; begin { nothing } end;
procedure TCustomHighlighter.BeforeSave; begin { nothing } end;
procedure TCustomHighlighter.Destroyed; begin { nothing } end;
procedure TCustomHighlighter.Modified; begin { Nothing } end;

function TCustomHighlighter._AddRef: Integer;
begin
  Result := -1;
end;

function TCustomHighlighter._Release: Integer;
begin
  Result := -1;
end;

function TCustomHighlighter.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface( IID, Obj ) then
    Result := s_ok
  else
    Result := e_NoInterface;
end;

procedure TCustomHighlighter.AddMessage(const sz: String);
var
  ms: IOTAMessageServices;
begin
  ms := GetIDEMessageService;
  ms.AddToolMessage( '', sz, FName, 0, 0 );
end;

constructor TCustomHighlighter.Create(const AName, AID: String);
begin
  inherited Create;
  FName := AName;
  FIDString := AID;
  InstallHighlighter( Self );
end;

destructor TCustomHighlighter.Destroy;
begin
  inherited Destroy;
end;

function TCustomHighlighter.FindHighlighter(const AID: String): IOTAHighlighter;
var
  i: Integer;
  hs: IOTAHighlightServices;
begin
  Result := nil;
  hs := GetIDEHighlighterServices;
  for i := 0 to hs.HighlighterCount - 1 do
    if hs.Highlighter[i].IDString = AID then begin
      Result := hs.Highlighter[i];
      Exit;
      end;
end;

function TCustomHighlighter.GetIDString: string;
begin
  Result := FIDString;
end;

function TCustomHighlighter.GetLineText(p: PChar; len: Integer): String;
begin
  SetLength( Result, len );
  Move( p^, Result[1], len );
end;

function TCustomHighlighter.GetName: string;
begin
  Result := FName;
end;

procedure TCustomHighlighter.InstallHighlighter;
var
  hs: IOTAHighlightServices;
begin
  assert( FHighlighterIndex = 0 );
  hs := GetIDEHighlighterServices;
  FHighlighterIndex := hs.AddHighlighter( Self );
end;

procedure TCustomHighlighter.RemoveHighlighter(const AID: String);
var
  i, delIndex: Integer;
  hs: IOTAHighlightServices;
begin
  delIndex := -1;
  hs := GetIDEHighlighterServices;
  for i := 0 to hs.HighlighterCount - 1 do
    if hs.Highlighter[i].IDString = AID then begin
      delIndex := i;
      break;
      end;
  assert( delIndex >= 0 );
  hs.RemoveHighlighter( delIndex );
end;

function TCustomHighlighter.GetHighlighterList: String;
var
  sz: String;
  i: Integer;
  hs: IOTAHighlightServices;
begin
  sz := '';
  hs := GetIDEHighlighterServices;
  for i := 0 to hs.HighlighterCount - 1 do
    sz := sz + Format( '%s, %s'#13#10, [
        hs.Highlighter[i].Name,
        hs.Highlighter[i].IDString
        ] );
  Result := sz;
end;

procedure TCustomHighlighter.Tokenize(StartClass: TOTALineClass;
  LineBuf: POTAEdChar; LineBufLen: TOTALineSize;
  HighlightCodes: POTASyntaxCode);
begin
  { nothing }
end;

function TCustomHighlighter.TokenizeLineClass(StartClass: TOTALineClass;
  LineBuf: POTAEdChar; LineBufLen: TOTALineSize): TOTALineClass;
begin
  Result := 0;
end;

{ TDelegateHighlighter }

constructor TDelegateHighlighter.Create(const AName, AID: String);
begin
  FDelegate := FindHighlighter( AID );
  assert( FDelegate <> nil );
  RemoveHighlighter( AID );
  inherited Create( AName, AID );
end;

destructor TDelegateHighlighter.Destroy;
begin
  try
    RemoveHighlighter( FIDString );
    InstallHighlighter( FDelegate );
    FDelegate := nil;
    inherited Destroy;
  except
    OutputDebugString( 'TDelegateHighlighter.Destroy failed' );
    end;
end;

procedure TDelegateHighlighter.Tokenize(StartClass: TOTALineClass;
  LineBuf: POTAEdChar; LineBufLen: TOTALineSize;
  HighlightCodes: POTASyntaxCode);
begin
  try
    FDelegate.Tokenize( StartClass, LineBuf, LineBufLen, HighlightCodes );
  except
    OutputDebugString( pchar('Tokenize: ' + Exception(ExceptObject).Message) );
    end;
end;

function TDelegateHighlighter.TokenizeLineClass(StartClass: TOTALineClass;
  LineBuf: POTAEdChar; LineBufLen: TOTALineSize): TOTALineClass;
begin
  try
    Result := FDelegate.TokenizeLineClass( StartClass, LineBuf, LineBufLen );
  except
    OutputDebugString( pchar('TokenizeLineClass: ' + Exception(ExceptObject).Message) );
    Result := 0;
    end;
end;

{ TAggregateHighlighter }

constructor TAggregateHighlighter.Create(const AName, AID, AAggregateID: String);
begin
  FAggregateID := AAggregateID;
  FAggregate := FindHighlighter( FAggregateID );
  assert( FAggregate <> nil );
  inherited Create( AName, AID );
end;

destructor TAggregateHighlighter.Destroy;
begin
  FAggregate := nil;
  inherited Destroy;
end;

procedure TAggregateHighlighter.Tokenize(StartClass: TOTALineClass;
  LineBuf: POTAEdChar; LineBufLen: TOTALineSize;
  HighlightCodes: POTASyntaxCode);
begin
  try
    FAggregate.Tokenize( StartClass, LineBuf, LineBufLen, HighlightCodes );
  except
    OutputDebugString( pchar('Tokenize: ' + Exception(ExceptObject).Message) );
    end;
end;

function TAggregateHighlighter.TokenizeLineClass(StartClass: TOTALineClass;
  LineBuf: POTAEdChar; LineBufLen: TOTALineSize): TOTALineClass;
begin
  try
    Result := FAggregate.TokenizeLineClass( StartClass, LineBuf, LineBufLen );
  except
    OutputDebugString( pchar('TokenizeLineClass: ' + Exception(ExceptObject).Message) );
    Result := 0;
    end;
end;

{ TIniHighlighter }

constructor TIniHighlighter.Create;
begin
//  inherited Create( 'INI Files', 'Borland.Highlighter.INI' ); // delegate
//  inherited Create( 'INI Files (RTS)', 'RTS.Highlighter.INI' ); // custom
  inherited Create( 'INI Files (RTS)', 'RTS.Highlighter.INI', 'Borland.Highlighter.INI' );
end;

procedure TIniHighlighter.Tokenize(StartClass: TOTALineClass;
  LineBuf: POTAEdChar; LineBufLen: TOTALineSize;
  HighlightCodes: POTASyntaxCode);
var
  i: Integer;
  underline, didEqual: Boolean;
begin
  {
    ACTION: this is just an example, the highlighting doesn't make much
            sense and the parsing is very simple
  }
  inherited Tokenize( StartClass, LineBuf, LineBufLen, HighlightCodes );
  underline := false; didEqual := false;
  for i := 0 to LineBufLen - 1 do begin
    case LineBuf^ of
      '[': begin
          if HighlightCodes^ <> atComment then
            underline := true;
          HighlightCodes^ := atReservedWord;
          end;
      ']': begin
          if HighlightCodes^ <> atComment then
            underline := false;
          HighlightCodes^ := atReservedWord;
          end;
      '=': begin
        if (HighlightCodes^ <> atComment) and not didEqual then begin
          didEqual := true;
          HighlightCodes^ := atSymbol;
          end;
        end;
      else begin
        if underline and (LineBuf^ <> ' ') then
          HighlightCodes^ := atScripts
        else if HighlightCodes^ <> atComment then
          if didEqual then
            HighlightCodes^ := atAttrValues
          else
            HighlightCodes^ := atAttrNames;
        end;
      end; { case }
    Inc( LineBuf ); Inc( HighlightCodes );
    end;
end;

initialization
  { nothing }
finalization
  Cleanup;
end.
